import numpy as np
import multiprocessing as mp
import argparse
import time
import sys
import math
import random
from kspeed.kspeedcluster import kspeedclient
from kspeed.utils.kspeed_util import size_dict, kspeed_adduser, kspeed_deluser, kspeed_new_file, kspeed_rm_file
import rediscluster

tw = 12
kspeedcli = None


def write_rate(case, sample_len, batch_size, cpl_array):
    with open(case + '_rate.log', 'w') as f:
        start_cpl_cnt = cpl_array.sum()
        while start_cpl_cnt != cpl_array.size:
            time.sleep(0.5)
            end_cpl_cnt = cpl_array.sum()
            cpl_cnt = end_cpl_cnt - start_cpl_cnt
            rate = sample_len * 4 * 8 * batch_size * cpl_cnt / 0.5
            f.write(case + ',' + str(rate))
            start_cpl_cnt = end_cpl_cnt
        f.write('\n')


def kspeed_get_bench(pid, args, kspeed_iplist_str, cost_array,
                     total_cost_array, cpl_array, test_key_array, start_pos,
                     nkey_proc):
    # kspeedcli = kspeedclient(str(kspeed_iplist_str),
    #                          args.username, args.password)
    ksdataset = kspeedcli.loadArray(args.filename, 1)

    idx_a = np.arange(args.iteration, dtype=np.int64)
    if args.shuffle:
        np.random.shuffle(idx_a)
    idx_a = idx_a % args.sample_num
    time_s0 = int(round(time.time() * 1000000))
    for idx in range(start_pos, nkey_proc):
        try:
            time_s = int(round(time.time() * 1000000))
            data_out = ksdataset.get(idx_a[idx])
            time_t = int(round(time.time() * 1000000))
            cost_array[idx] = time_t - time_s
            cpl_array[idx] = 1
        except Exception as e:
            print("kspeed get error:", e)
            sys.exit(-1)
    time_e0 = int(round(time.time() * 1000000))
    total_cost_array[pid] = time_e0 - time_s0


def kspeed_getbatch_bench(pid, args, kspeed_iplist_str, cost_array,
                          total_cost_array, cpl_array, test_key_array,
                          start_pos, nkey_proc):
    # kspeedcli = kspeedclient(str(kspeed_iplist_str),
    #                          args.username, args.password)
    ksdataset = kspeedcli.loadArray(args.filename, args.batch_size)

    idx_a = np.arange(args.iteration, dtype=np.int64)
    if args.shuffle:
        np.random.shuffle(idx_a)
    idx_a = idx_a % args.sample_num
    time_s0 = int(round(time.time() * 1000000))
    for idx in range(start_pos, nkey_proc, args.batch_size):
        if idx + args.batch_size > args.iteration:
            batch_size = args.iteration - idx
        else:
            batch_size = args.batch_size

        try:
            idx_b = idx_a[idx:idx + batch_size]
            time_s = int(round(time.time() * 1000000))
            data_out = ksdataset.getbatch(idx_b, batch_size)
            time_t = int(round(time.time() * 1000000))
            cost_array[idx] = time_t - time_s
            cpl_array[idx] = 1
        except Exception as e:
            print("kspeed getbatch error:", e)
            sys.exit(-1)
    time_e0 = int(round(time.time() * 1000000))
    total_cost_array[pid] = time_e0 - time_s0


def redis_get_bench(pid, args, startup_nodes, cost_array, total_cost_array,
                    cpl_array, test_key_array, start_pos, nkey_proc):
    r = rediscluster.RedisCluster(startup_nodes=startup_nodes)

    time_s0 = int(round(time.time() * 1000000))
    for idx in range(start_pos, nkey_proc):
        try:
            time_s = int(round(time.time() * 1000000))
            data_out = r.get(idx)
            data_out_t = np.fromstring(data_out)
            time_t = int(round(time.time() * 1000000))
            cost_array[idx] = time_t - time_s
            cpl_array[idx] = 1
            if data_out is None:
                print("redis get null")
        except Exception as e:
            print("redis get error:", e)
            sys.exit(-1)
    time_e0 = int(round(time.time() * 1000000))
    total_cost_array[pid] = time_e0 - time_s0


def kspeed_bench(args, kspeed_iplist_str):
    # prepare test keys
    kspeed_new_file(args.username, args.password, kspeed_iplist_str,
                    args.sample_num, args.sample_len, args.filename, True)
    if args.batch_mode == None:
        nkey = args.sample_num if args.iteration > args.sample_num else args.iteration
    else:
        nkey_i = args.iteration * args.batch_size
        nkey = args.sample_num if nkey_i > args.sample_num else nkey_i

    if nkey > args.sample_num:
        get_keys_list = [i for i in range(args.sample_num)]
        for i in range(args.sample_num, args.iteration):
            get_keys_list.append(get_keys_list[i % args.sample_num])
        get_keys = np.array(get_keys_list)
    else:
        get_keys = np.arange(args.iteration)

    if args.shuffle:
        np.random.shuffle(get_keys)

    bench_num_per_size = [1, 2, 4, 8, 16, 32
                          ] if args.process_num == 255 else [args.process_num]
    test_key_array = mp.Manager().Array('i', get_keys)
    cost_array = mp.Manager().Array('i', np.ones(nkey, dtype=np.int64))
    for n in bench_num_per_size:
        progs = []
        keys_per_proc = math.ceil(nkey / n)
        total_cost_array = mp.Manager().Array('i', np.ones(n, dtype=np.int64))
        cpl_array = mp.Manager().Array('i', np.zeros(nkey, dtype=np.int64))
        for i in range(n):
            start_pos = i * keys_per_proc
            end_pos = (i + 1) * keys_per_proc
            if end_pos > args.iteration:
                nkey_proc = end_pos - start_pos
            else:
                nkey_proc = keys_per_proc
            if args.batch_mode == None:
                p = mp.Process(target=kspeed_get_bench,
                               args=(i, args, kspeed_iplist_str, cost_array,
                                     total_cost_array, cpl_array,
                                     test_key_array, start_pos, nkey_proc))
            else:
                p = mp.Process(target=kspeed_getbatch_bench,
                               args=(i, args, kspeed_iplist_str, cost_array,
                                     total_cost_array, cpl_array,
                                     test_key_array, start_pos, nkey_proc))
            progs.append(p)
            p.start()
        if args.batch_mode == None:
            write_rate('s_' + str(args.sample_len) + 'p_' + str(n),
                       args.sample_len, 1, cpl_array)
        else:
            write_rate('batch_s_' + str(args.sample_len) + 'p_' + str(n),
                       args.sample_len, args.batch_size, cpl_array)

        for j in progs:
            j.join()

        max_read_time = np.max(total_cost_array)
        avg_bw = args.iteration * args.sample_len * 4 * 8 / max_read_time
        avg_tps = args.iteration * 1e6 / max_read_time
        print(
            str(args.process_num).ljust(tw) + str(args.sample_len).ljust(tw) +
            str(args.batch_size).center(tw) + str(args.iteration).center(tw) +
            str(np.mean(cost_array)).center(tw) +
            str(np.min(cost_array)).center(tw) +
            str(np.max(cost_array)).center(tw) +
            str(np.percentile(cost_array, 50)).center(tw) +
            str(np.percentile(cost_array, 90)).center(tw) +
            str(np.percentile(cost_array, 99)).center(tw) +
            str(avg_bw).center(tw + 10) + str(avg_tps).center(tw + 10))

    kspeed_rm_file(args.username, args.password, kspeed_iplist_str,
                   args.filename, True)


def redis_bench(args, startup_nodes):
    # prepare test keys
    # nkey_i = args.sample_num if args.iteration > args.sample_num else args.iteration
    nkey_i = 16384 if args.iteration > 16384 else args.iteration
    nkey_set = nkey_i if nkey_i > 4096 else 4096
    redis_set_keys = np.random.randint(0,
                                       sys.maxsize,
                                       nkey_set,
                                       dtype=np.int64)
    r = rediscluster.RedisCluster(startup_nodes=startup_nodes)
    value = np.ones(args.sample_len).tobytes()
    for key in redis_set_keys:
        r.set(key, value.tobytes())

    get_keys_list = []
    for i in range(args.iteration):
        get_keys_list.append(redis_set_keys[i % nkey_set])
    get_keys = np.array(get_keys_list)

    test_key_array = mp.Manager().Array('i', get_keys)
    cost_array = mp.Manager().Array('i', np.ones(args.iteration,
                                                 dtype=np.int64))

    bench_num_per_size = [1, 2, 4, 8, 16, 32
                          ] if args.process_num == 255 else [args.process_num]
    for n in bench_num_per_size:
        progs = []
        keys_per_proc = math.ceil(args.iteration / n)
        cpl_array = mp.Manager().Array(
            'i', np.zeros(args.iteration, dtype=np.int64))
        total_cost_array = mp.Manager().Array(
            'i', np.ones(args.iteration, dtype=np.int64))
        for i in range(n):
            start_pos = i * keys_per_proc
            end_pos = (i + 1) * keys_per_proc
            if end_pos > args.iteration:
                nkey_proc = end_pos - start_pos
            else:
                nkey_proc = keys_per_proc
            p = mp.Process(target=redis_get_bench,
                           args=(i, args, startup_nodes, cost_array,
                                 total_cost_array, cpl_array, test_key_array,
                                 start_pos, nkey_proc))
            progs.append(p)
            p.start()

        write_rate('s_' + str(args.sample_len) + 'p_' + str(n),
                   args.sample_len, 1, cpl_array)

        for j in progs:
            j.join()

        max_read_time = np.max(total_cost_array)
        avg_bw = args.iteration * args.sample_len * 4 * 8 / max_read_time
        avg_tps = args.iteration * 1e6 / max_read_time
        print(
            str(args.process_num).ljust(tw) + str(args.sample_len).ljust(tw) +
            str(args.batch_size).center(tw) + str(args.iteration).center(tw) +
            str(np.mean(cost_array)).center(tw) +
            str(np.min(cost_array)).center(tw) +
            str(np.max(cost_array)).center(tw) +
            str(np.percentile(cost_array, 50)).center(tw) +
            str(np.percentile(cost_array, 90)).center(tw) +
            str(np.percentile(cost_array, 99)).center(tw) +
            str(avg_bw).center(tw + 10) + str(avg_tps).center(tw + 10))


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("cache_type", type=str, help="specify kspeed or redis")
    parser.add_argument('-u',
                        "--username",
                        type=str,
                        dest="username",
                        default='alitest_bench',
                        help="specify the username to login")
    parser.add_argument('-w',
                        "--password",
                        type=str,
                        dest="password",
                        default='alitest_bench',
                        help="specify the password to login")
    parser.add_argument('-i',
                        "--iplist",
                        dest="iplist",
                        type=str,
                        default='./iplist',
                        help="specify the kspeed servers iplist ")
    parser.add_argument('-f',
                        "--filename",
                        dest="filename",
                        type=str,
                        default='default',
                        help="specify the file name to read")
    parser.add_argument('-b',
                        "--batch_size",
                        dest="batch_size",
                        type=int,
                        default=256,
                        help="specify the batch_size of data loading")
    parser.add_argument('-g',
                        "--to_gpu",
                        dest="to_gpu",
                        type=int,
                        default=1000,
                        help="specify the gpu_no which the data load to ")
    parser.add_argument(
        '-l',
        "--sample_len",
        dest="sample_len",
        type=int,
        default=0,
        help="specify the sample length which will be read from kspeed ")
    parser.add_argument(
        '-d',
        "--sample_num",
        dest="sample_num",
        type=int,
        default=0,
        help="specify the sample num which will be read from kspeed ")
    parser.add_argument(
        '-n',
        "--process_num",
        dest="process_num",
        type=int,
        default=1,
        help="specify the sample num which will be read from kspeed ")
    parser.add_argument('-t',
                        "--iteration",
                        dest="iteration",
                        type=int,
                        default=10000,
                        help="specify the iteration to run test")
    parser.add_argument('-m',
                        "--batch_mode",
                        dest="batch_mode",
                        type=str,
                        default='batch',
                        choices=["batch", "None"],
                        help="specify the mode using to load a batch of data")
    parser.add_argument('-s',
                        "--shuffle",
                        dest="shuffle",
                        type=bool,
                        default=True,
                        help="specify True to have the data shuffled")

    args = parser.parse_args()
    print("COMMAND: {}".format(args))
    return args


def main():
    args = parse_args()

    if args.sample_len == 0:
        sample_len_idx = [i for i in size_dict][2:16]
        sample_len_test = [int(size_dict[i] / 4) for i in sample_len_idx]
    else:
        sample_len_idx = [args.sample_len]
        sample_len_test = [args.sample_len / 4]
    if args.cache_type == 'kspeed':
        with open(args.iplist, 'r') as f:
            kspeed_iplist_str = ' '.join(f.read().split())
        kspeed_adduser(kspeed_iplist_str, args.username, args.password, True)

    print("\nnprocess".ljust(tw) + "sample_len".ljust(tw) +
          "batch_size".center(tw) + "iteration".center(tw) +
          "avg(us)".center(tw) + "min(us)".center(tw) + "max(us)".center(tw) +
          "p50(us)".center(tw) + "p90(us)".center(tw) + "p99(us)".center(tw) +
          "avg_bw".center(tw + 10) + "avg_tps".center(tw + 10))
    for i in range(len(sample_len_test)):
        args.sample_len = sample_len_test[i]
        args.filename = "py-bench-" + str(sample_len_idx[i])
        args.sample_num = 4 * int(512 * 1024 * 1024 / args.sample_len / 4)
        if args.cache_type == 'kspeed':
            kspeed_bench(args, kspeed_iplist_str)
        else:
            startup_nodes = [{
                "host": "11.22.33.20",
                "port": "9376"
            }, {
                "host": "11.22.33.140",
                "port": "9377"
            }, {
                "host": "11.22.33.154",
                "port": "9378"
            }]
            redis_bench(args, startup_nodes)

    if args.cache_type != 'redis':
        kspeed_deluser(args.username, kspeed_iplist_str, True)


if __name__ == '__main__':
    main()
