import time
import os
import argparse
import pandas as pd
import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from torch.distributed import get_rank, get_world_size
import torchvision.transforms as transforms
from PIL import Image
import torchvision.datasets as datasets

from kspeed.utils.data.kspeeddataset import KSpeedFolder, KSpeedImageFolder

import warnings

warnings.filterwarnings('ignore')

barrier_tensor = torch.ones(2, 16)


def setup_training(local_rank, ngpus, backend):
    # initialize the process group
    global barrier_tensor
    if backend != 'gloo':
        torch.cuda.set_device(local_rank)
        barrier_tensor = torch.ones(2, 16).cuda()

    if os.getenv('MASTER_ADDR') is None:
        os.environ['MASTER_ADDR'] = '127.0.0.1'
    if os.getenv('MASTER_PORT') is None:
        os.environ['MASTER_PORT'] = '9785'
    if os.getenv('RANK') is None:
        os.environ['RANK'] = '0'
    if os.getenv('WORLD_SIZE') is None:
        os.environ['WORLD_SIZE'] = '1'
    master_ip = os.environ['MASTER_ADDR']
    master_port = os.environ['MASTER_PORT']
    worker_size = int(os.environ['WORLD_SIZE'])
    worker_rank = int(os.environ['RANK'])
    ngpus_per_worker = ngpus  #int(os.environ['KUBERNETES_CONTAINER_RESOURCE_GPU'])
    world_size = worker_size * ngpus_per_worker
    world_rank = worker_rank * ngpus_per_worker + local_rank
    init_method = 'tcp://' + master_ip + ':' + master_port
    print(
        f"[{os.getpid()}] Initializing process group with: master_ip:{master_ip}, master_port:{master_port}, worker_size:{worker_size},worker_rank:{worker_rank}, ngpu_per_worker:{ngpus_per_worker}, init_method:{init_method}, world_size:{world_size}, world_rank:{world_rank}, local_rank:{local_rank}"
    )
    dist.init_process_group(backend=backend,
                            world_size=world_size,
                            rank=world_rank)
    # print(
    #     f"[{os.getpid()}] : world_rank{world_rank} start all_reduce for init_process_group"
    # )
    dist.all_reduce(barrier_tensor, op=dist.ReduceOp.SUM, async_op=False)
    dist.barrier()
    # print(
    #     f"[{os.getpid()}] : world_rank{world_rank} finished all_reduce for init_process_group"
    # )


def expand(num_classes, dtype, tensor):
    e = torch.zeros(
        tensor.size(0), num_classes, dtype=dtype, device=torch.device("cuda")
    )
    e = e.scatter(1, tensor.unsqueeze(1), 1.0)
    return e


class PrefetchedWrapper(object):
    def prefetched_loader(loader, num_classes, one_hot):
        mean = (
            torch.tensor([0.485 * 255, 0.456 * 255, 0.406 * 255])
            .cuda()
            .view(1, 3, 1, 1)
        )
        std = (
            torch.tensor([0.229 * 255, 0.224 * 255, 0.225 * 255])
            .cuda()
            .view(1, 3, 1, 1)
        )

        stream = torch.cuda.Stream()
        first = True

        for next_input, next_target in loader:
            with torch.cuda.stream(stream):
                next_input = next_input.cuda(non_blocking=True)
                next_target = next_target.cuda(non_blocking=True)
                next_input = next_input.float()
                if one_hot:
                    next_target = expand(num_classes, torch.float, next_target)

                next_input = next_input.sub_(mean).div_(std)

            if not first:
                yield input, target
            else:
                first = False

            torch.cuda.current_stream().wait_stream(stream)
            input = next_input
            target = next_target

        yield input, target

    def __init__(self, dataloader, start_epoch, num_classes, one_hot):
        self.dataloader = dataloader
        self.epoch = start_epoch
        self.one_hot = one_hot
        self.num_classes = num_classes

    def __iter__(self):
        if self.dataloader.sampler is not None and isinstance(
            self.dataloader.sampler, torch.utils.data.distributed.DistributedSampler
        ):

            self.dataloader.sampler.set_epoch(self.epoch)
        self.epoch += 1
        return PrefetchedWrapper.prefetched_loader(
            self.dataloader, self.num_classes, self.one_hot
        )

    def __len__(self):
        return len(self.dataloader)


#----------------------------training part ----------------------------
def data_loading(dl, args):
    world_rank = get_rank()
    world_size = get_world_size()
    perf_metrics = []
    path = 'dl_benchmark_data'
    if world_rank % 8 == 0:
        isExist = os.path.exists(path)
        if not isExist:
            os.makedirs(path)

    global barrier_tensor
    for e in range(args.epoch_num + 1):
        epoch_metrics = []
        data_time_s = int(round(time.time() * 1000000))
        for i, data in enumerate(dl):
            if len(data) == 2:
                batch, target = data
            else:
                batch = data
            data_time_e = int(round(time.time() * 1000000))
            data_time = data_time_e - data_time_s
            dist.barrier()
            sync_time_e = int(round(time.time() * 1000000))
            sync_time = sync_time_e - data_time_e
            dl_time = sync_time_e - data_time_s
            data_time_s = int(round(time.time() * 1000000))
            iter_metrics = {'epoch': e, 'iter': i, 'data_time': data_time, 'sync_time': sync_time, 'dl_time': dl_time}
            epoch_metrics.append(iter_metrics)
            if i % args.print_freq == 0 and world_rank % 8 == 0:
                print(
                    "rank {} epoch {} iteration {} : batch.shape = {}, data_time = {}us, sync_time ={}us, dl_time = {}us"
                    .format(world_rank, e, i, batch.shape, data_time, sync_time, dl_time))
        metrics = pd.DataFrame(epoch_metrics)
        perf_metrics.append(metrics)
        data_time_series = metrics["data_time"]
        sync_time_series = metrics['sync_time']
        dl_time_series = metrics['dl_time']
    #     print(
    #         f'rank {world_rank} epoch [{e}]: data_time mean = {data_time_series.mean()} us, std = {data_time_series.std()}, sync_time ={sync_time_series.mean()} us,  std = {sync_time_series.std()}, dl_time = {dl_time_series.mean()} us,  std = {dl_time_series.std()}')
    # print(
    #     f"[{os.getpid()}] : world_rank{world_rank} start all_reduce after dataloading"
    # )
    dist.all_reduce(barrier_tensor, op=dist.ReduceOp.SUM, async_op=False)
    dist.barrier()
    # print(
    #     f"[{os.getpid()}] : world_rank{world_rank} finished all_reduce after dataloading"
    # )
    filename = f'{path}/dl_benchmark_{args.data_source}_w{world_size}_r{world_rank}.csv'
    for e in range(args.epoch_num + 1):
        data = perf_metrics[e]
        if e == 0:
            data.to_csv(filename, mode='w')
        else:
            data.to_csv(filename, mode='a')
            if e == 2:
                df = data
            elif e > 2:
                df = pd.concat([df, data])
    data_time_series = df["data_time"]
    sync_time_series = df['sync_time']
    dl_time_series = df['dl_time']
    if world_rank == 0:
        print(
            f'world_size: {world_size} summary []: data_time mean = {data_time_series.mean()} us, std = {data_time_series.std()}, sync_time ={sync_time_series.mean()} us,  std = {sync_time_series.std()}, dl_time = {dl_time_series.mean()} us,  std = {dl_time_series.std()}')


def collate_fake(batch):
    if (len(batch[0]) == 1):
        return batch
    imgs = [img[0] for img in batch]
    targets = torch.tensor([target[1] for target in batch], dtype=torch.int64)
    tensor = torch.zeros((len(imgs), 3, 244, 244), dtype=torch.uint8)
    return tensor, targets


def main_worker(gpuid, args):
    setup_training(gpuid, args.ngpus, args.backend)
    #training dataset
    interpolation = "bilinear"
    interpolation = {"bicubic": Image.BICUBIC, "bilinear": Image.BILINEAR}[
        interpolation
    ]
    transforms_list = [
        transforms.RandomResizedCrop(244, interpolation=interpolation),
        transforms.RandomHorizontalFlip(),
    ]
    collate_fn = collate_fake
    if args.data_source == 'kspeed':
        train_ds = KSpeedImageFolder(args.root, args.objname, args.dlworker,
                                     args.iplist, args.username,
                                     args.password)
    elif args.data_source == 'kspeed_with_preprocess':
        train_ds = KSpeedImageFolder(args.root, args.objname, args.dlworker,
                                     args.iplist, args.username,
                                     args.password, transforms.Compose(transforms_list))
    elif args.data_source == 'fs_with_preprocess':
        train_ds = datasets.ImageFolder(args.root, transforms.Compose(transforms_list))
    elif args.data_source != 'fs':
        train_ds = datasets.ImageFolder(args.root)
    else:
        raise ValueError(f"Unsupported data_source {args.data_source}, exit!")

    #training dataloader
    if torch.distributed.is_initialized():
        train_sampler = torch.utils.data.distributed.DistributedSampler(
            train_ds)
    else:
        train_sampler = None

    dataloader = torch.utils.data.DataLoader(train_ds,
                                             batch_size=args.batch_size,
                                             sampler=train_sampler,
                                             shuffle=(train_sampler is None),
                                             collate_fn=collate_fn,
                                             pin_memory=True,
                                             drop_last=True,
                                             persistent_workers=True,
                                             prefetch_factor=4,
                                             num_workers=args.dlworker)

    #data loading test
    data_loading(dataloader, args)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i',
                        '--iplist',
                        dest="iplist",
                        default="./iplist",
                        type=str,
                        help='iplist of kspeed server')
    parser.add_argument('-b',
                        '--batch_size',
                        dest="batch_size",
                        default=256,
                        type=int,
                        help='batch_size for distributed training')
    parser.add_argument('-n',
                        '--dlworker',
                        dest="dlworker",
                        default=10,
                        type=int,
                        help='dlworker for dataloader')
    parser.add_argument('-e',
                        '--epoch_num',
                        dest="epoch_num",
                        default=1,
                        type=int,
                        help='epoch_num for distributed training')
    parser.add_argument('-u',
                        "--username",
                        type=str,
                        dest="username",
                        default='admin',
                        help="specify the username to login")
    parser.add_argument('-w',
                        "--password",
                        type=str,
                        dest="password",
                        default='admin',
                        help="specify the password to login")
    parser.add_argument('-g',
                        '--ngpus',
                        dest="ngpus",
                        default=1,
                        type=int,
                        help='gpu num to use ')
    parser.add_argument('--backend',
                        dest="backend",
                        default="nccl",
                        type=str,
                        choices=["gloo", "nccl", "accl"],
                        help='communication backend to use')
    parser.add_argument('--print_freq',
                        dest="print_freq",
                        default=10000000,
                        type=int,
                        help='print freq')
    parser.add_argument(
        '-s',
        '--data_source',
        dest="data_source",
        default='kspeed',
        type=str,
        help='specify the source of data input')
    parser.add_argument('-o',
                        "--objname",
                        dest="objname",
                        type=str,
                        help="specify the kspeed object name to test")
    parser.add_argument('-r',
                        "--root",
                        dest="root",
                        type=str,
                        help="specify the root path of dataset")

    args = parser.parse_args()
    print("COMMAND: {}".format(args))
    return args


def main():
    args = parse_args()
    if os.getenv('gpus') is not None:
        args.ngpus = int(os.environ['gpus'])
    mp.spawn(main_worker, nprocs=args.ngpus, args=(args, ))


if __name__ == '__main__':
    main()

