import sys
import time
import os
import argparse
import subprocess

import torch
import datetime
import math
import numpy as np
import numpy.testing as npt
import torch.multiprocessing as mp
from torch.utils.data import BatchSampler, RandomSampler, SequentialSampler
import datetime
import h5py
from kspeed.kspeedcluster import kspeedclient
from kspeed.utils.kspeed_util import size_dict, kspeed_adduser, kspeed_deluser, kspeed_new_file, kspeed_rm_file, get_kspeed_len, kspeed_cache_file
from kspeed.utils.data.dataloader import KSpeedDataLoader
from kspeed.utils.data.dataset import KSpeedDataset


def setup_training(args):
    if os.getenv('MASTER_ADDR') == None:
        os.environ['MASTER_ADDR'] = '127.0.0.1'
    if os.getenv('MASTER_PORT') == None:
        os.environ['MASTER_PORT'] = '9785'
    if os.getenv('RANK') == None:
        os.environ['RANK'] = '0'
    if os.getenv('WORLD_SIZE') == None:
        os.environ['WORLD_SIZE'] = '1'
    master_ip = os.getenv('MASTER_ADDR')
    master_port = os.getenv('MASTER_PORT')
    rank = int(os.getenv('RANK'))
    world_size = int(os.getenv('WORLD_SIZE'))
    print(
        f"[{os.getpid()}] Initializing process group with: master_ip:{master_ip},master_port:{master_port},world_size:{world_size}, world_rank:{rank}, local_rank: {args.local_rank}"
    )
    torch.distributed.init_process_group(backend='nccl', init_method='env://')

    if args.gpu >= 0:
        device = torch.device('cuda:' + str(args.gpu))
        torch.cuda.set_device(args.gpu)


import warnings

warnings.filterwarnings('ignore')

server_iplist_str = ' '


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-r',
                        '--local_rank',
                        dest="local_rank",
                        default=0,
                        type=int,
                        help='node local rank for distributed training')
    parser.add_argument('-i',
                        '--iplist',
                        dest="iplist",
                        default="./iplist",
                        type=str,
                        help='iplist of kspeed server')
    parser.add_argument('-b',
                        '--batch_size',
                        dest="batch_size",
                        default=200,
                        type=int,
                        help='batch_size for distributed training')
    parser.add_argument('-n',
                        '--dlworker',
                        dest="dlworker",
                        default=1,
                        type=int,
                        help='dlworker for dataloader')
    parser.add_argument('-e',
                        '--epoch_num',
                        dest="epoch_num",
                        default=1,
                        type=int,
                        help='epoch_num for distributed training')
    parser.add_argument('-u',
                        "--username",
                        type=str,
                        dest="username",
                        default='alitest_bench',
                        help="specify the username to login")
    parser.add_argument('-w',
                        "--password",
                        type=str,
                        dest="password",
                        default='alitest_bench',
                        help="specify the password to login")
    parser.add_argument('-m',
                        '--batch_mode',
                        dest="batch_mode",
                        default='batch',
                        type=str,
                        choices=["batch", "None"],
                        help='specify the mode using to load a batch of data ')
    parser.add_argument(
        '-g',
        '--gpu',
        dest="gpu",
        default=-1,
        type=int,
        help='specify the gpu number which batch of data load to')
    parser.add_argument('-k',
                        '--dl_backend',
                        dest="dl_backend",
                        default='kspeed',
                        type=str,
                        choices=["kspeed", "pytorch"],
                        help='specify the source of data')
    parser.add_argument('-f',
                        "--filename",
                        dest="filename",
                        type=str,
                        help="specify the kspeed file name to test")
    parser.add_argument('-sf',
                        "--src_filename",
                        dest="src_filename",
                        type=str,
                        help="specify the file name to test")
    parser.add_argument("--shuffle",
                        dest="shuffle",
                        type=bool,
                        default=True,
                        help="specify True to have the data shuffled")

    args = parser.parse_args()
    print("COMMAND: {}".format(args))

    with open(args.iplist, 'r') as f:
        server_iplist_str = ' '.join(f.read().split())

    setup_training(args)

    rm_file_flag = False
    if args.filename is None:
        rm_file_flag = True
        kspeed_adduser(server_iplist_str, args.username, args.password,
                       args.dlworker == 0)
        if 'npy' in args.src_filename:
            args.filename = args.src_filename
            kspeed_cache_file(args.username, args.password, server_iplist_str,
                              args.src_filename, args.dlworker == 0)
        elif '.h5' in args.src_filename:
            subprocess.call("kspeedcmd hdf5_tokspeed -f {} -u {} -w {}".format(
                args.src_filename, args.username, args.password),
                            shell=True)
            with h5py.File(args.src_filename, 'r') as f:
                dataset_list = [key for key in f.keys()]
                args.filename = dataset_list[0]
        else:
            print(
                "Unsupported file type, while only npy and hdf5 are supported")

    #training dataset
    if args.batch_mode == 'None':
        train_ds = KSpeedDataset(args.username, args.password,
                                 server_iplist_str, args.filename, 1,
                                 args.dlworker == 0)
    else:
        train_ds = KSpeedDataset(args.username, args.password,
                                 server_iplist_str, args.filename,
                                 args.batch_size, args.dlworker == 0)

    #training dataloader
    if args.batch_mode == 'None':
        dataloader = torch.utils.data.DataLoader(train_ds,
                                                 batch_size=args.batch_size,
                                                 shuffle=True,
                                                 num_workers=args.dlworker)
    else:
        if args.shuffle:
            batch_sampler_tr = BatchSampler(RandomSampler(train_ds),
                                            batch_size=args.batch_size,
                                            drop_last=True)
        else:
            batch_sampler_tr = BatchSampler(SequentialSampler(train_ds),
                                            batch_size=args.batch_size,
                                            drop_last=True)

        if args.dl_backend == 'kspeed':
            dataloader = KSpeedDataLoader(train_ds,
                                          batch_sampler=batch_sampler_tr,
                                          batch_size=args.batch_size,
                                          num_workers=args.dlworker,
                                          pin_memory=args.gpu >= 0)
        else:
            dataloader = torch.utils.data.DataLoader(train_ds,
                                                     sampler=batch_sampler_tr,
                                                     batch_size=None,
                                                     num_workers=args.dlworker)

    for e in range(args.epoch_num):
        for i, batch in enumerate(dataloader):
            if 'npy' in args.src_filename:
                dataset_npy = np.load(args.src_filename, mmap_mode='r')
                src_data = torch.from_numpy(
                    dataset_npy[i * args.batch_size:i * args.batch_size +
                                len(batch)])
                torch.equal(batch, src_data)
                print(
                    "pytorch-kspeed-data-check-{} result pass: batch.shape = {}"
                    .format(i, batch.shape))
            else:  #hdf5
                with h5py.File(args.src_filename, 'r') as f:
                    dataset_h5 = f[args.filename]
                    src_numpy = dataset_h5[i * args.batch_size:i *
                                           args.batch_size + len(batch)]
                    src_data = torch.from_numpy(
                        dataset_h5[i * args.batch_size:i * args.batch_size +
                                   len(batch)])
                    torch.equal(batch, src_data)
                    print(
                        "pytorch-kspeed-data-check-{} result pass: batch.shape = {}"
                        .format(i, batch.shape))

    if rm_file_flag:
        kspeed_rm_file(args.username, args.password, server_iplist_str,
                       args.filename, args.dlworker == 0)
        kspeed_deluser(args.username, server_iplist_str, args.dlworker == 0)


if __name__ == '__main__':
    print("######################################################")
    print(
        "EXEC CMD : python -m torch.distributed.launch --nproc_per_node=1 kspeed-pytorch-data-check.py"
    )
    print("######################################################")
    main()
