import sys
import numpy as np
import numpy.testing as npt
import multiprocessing as mp
import progressbar

import click
from kspeed.kspeedcmd.kspeedcmd import pass_config
from kspeed.utils.kspeed_util import get_kspeed_len, kspeed_new_file


@click.command("new", short_help="New an empty array in KSpeed.")
@click.option(
    '-s',
    '--shape',
    type=str,
    required=True,
    help="Specify the shape of dataset to generate, e.g. \"4096, 1024\".")
@click.option("-o",
              "--objname",
              default="default_array",
              help="Specify the object name to new in KSpeed.")
@click.option("-t",
              "--dtype",
              type=str,
              default='float32',
              help="Specify the data type of dataset to generate.")
@pass_config
def new(ctx, objname, shape, dtype):
    """New an empty array in KSpeed."""
    ctx.kspeedcli.newArray(np.array(tuple(eval(shape))), 2, dtype, objname)


@click.command("set", short_help="Set an item to array in KSpeed.")
@click.option(
    '-s',
    '--shape',
    type=str,
    required=True,
    help="Specify the shape of dataset to generate, e.g. \"4096, 1024\".")
@click.option("-o",
              "--objname",
              default="default_array",
              help="Specify the object name to new in KSpeed.")
@click.option("-t",
              "--dtype",
              type=str,
              default='float32',
              help="Specify the data type of dataset to generate.")
@click.option('-n',
              '--process_num',
              type=int,
              default=1,
              help="Specify number of process to check.")
@pass_config
def set(ctx, objname, shape, dtype, process_num):
    """New an empty array in KSpeed."""
    shape_t = tuple(eval(shape))
    if process_num == 1:
        ctx.kspeedcli.newArray(np.array(shape_t), 2, dtype, objname)
        _array_set(0, ctx.kspeedcli, objname, dtype, 0, shape_t[0])
        return

    # with open(ctx.iplist, 'r') as f:
    #     iplist_str = ' '.join(f.read().split())
    # kspeed_new_file(iplist_str, shape, dtype, dest, False, ctx.username,
    #                 ctx.password)
    p = mp.Process(target=_new, args=(ctx.kspeedcli, shape_t, objname, dtype))
    p.start()
    p.join()

    idx_per_process = int(shape_t[0] / process_num)
    proc = []
    queue = mp.Queue()
    for i in range(process_num - 1):
        p = mp.Process(target=_array_set,
                       args=(i, ctx.kspeedcli, objname, dtype,
                             i * idx_per_process, (i + 1) * idx_per_process,
                             queue))
        p.start()
        proc.append(p)
    p = mp.Process(target=_array_set,
                   args=(process_num - 1, ctx.kspeedcli, objname, dtype,
                         (process_num - 1) * idx_per_process, shape_t[0],
                         queue))
    p.start()
    proc.append(p)

    for i in progressbar.progressbar(range(shape_t[0]),
                                     prefix='Set processing:'):
        queue.get()

    for i in proc:
        i.join()


def _new(kspeedcli, shape_t, dest, dtype):
    kspeedcli.newArray(np.array(shape_t), 2, dtype, dest)


def _array_set(pid, kspeedcli, dest, dtype, start, end, queue=None):
    dataset = kspeedcli.loadArray(dest)
    sample_bytelen = dataset.sample_len()
    if queue is None:
        bar = progressbar.ProgressBar(prefix='Set processing:',
                                      max_value=dataset.len()).start()

    for idx in range(start, end):
        idx_i = idx * 10
        value = idx_i.to_bytes(length=sample_bytelen,
                               byteorder='big',
                               signed=False)
        value_a = np.frombuffer(value, dtype=dtype)
        try:
            dataset.set(idx, value_a)
        except Exception as e:
            print(f'kspeed set idx = {idx} failed, {e}')
            sys.exit(-1)

        if queue is None:
            bar.update(idx)
        else:
            queue.put(idx)