import os
import progressbar
import multiprocessing as mp

from kspeed.kspeedcmd.kspeedcmd import pass_config
import click
import numpy.testing as npt
import numpy as np


@click.command("fileset_size", short_help="Get fileset size.")
@click.option('-o',
              '--objname',
              required=True,
              type=str,
              help="Specify the object name in KSpeed.")
@pass_config
def fileset_size(ctx, objname):
    """Get fileset size."""
    dataset = ctx.kspeedcli.loadFileSet(objname)
    click.secho(f"{objname}'s size = {dataset.size()}!", fg='red', bg='black')


@click.command("fileset_preload", short_help="Preload fileset to KSpeed.")
@click.option('-r',
              '--root',
              required=True,
              type=click.Path(exists=True),
              help="Specify the root path of fileset to preload to KSpeed.")
@click.option('-o',
              '--objname',
              type=str,
              help="Specify the object name in KSpeed.")
@click.option('-m',
              '--mode',
              default='w',
              type=click.Choice(['w', 'a']),
              help="Specify the mode to preload.")
@pass_config
def fileset_preload(ctx, root, objname, mode):
    """Preload fileset to KSpeed."""
    if root == "":
        click.secho("bad root path!", fg='red', bg='black')
        return
    root = _parse_root(root)
    ctx.kspeedcli.preloadFileSet(root, objname, mode)


@click.command("fileset_check",
               short_help="check the fileset preloaded to KSpeed.")
@click.option('-r',
              '--root',
              required=True,
              type=click.Path(exists=True),
              help="Specify the root path of fileset to check.")
@click.option('-o',
              '--objname',
              required=True,
              help="Specify the object name to check in KSpeed.")
@click.option('-n',
              "process_num",
              default=10,
              type=int,
              help="Specify the process num used to check.")
@click.option('-d',
              '--device',
              type=click.Choice(['CPU', 'GPU']),
              default='CPU',
              help="Specify the device to check in KSpeed.")
@click.option('-t',
              '--typer',
              type=click.Choice(['WAV', 'JPEG']),
              default='JPEG',
              help="Specify the class of the object to check in KSpeed.")
@pass_config
def fileset_check(ctx, root, objname, process_num, device, typer):
    """check the fileset preloaded to KSpeed."""
    if root == "":
        click.secho("bad root path!", fg='red', bg='black')
        return
    root = _parse_root(root)
    pnum = process_num
    try:
        fileset = _read_filelist(root)
        click.secho("fileset found!", fg='green', bg='black')
    except IOError:
        click.secho("No fileset found, traversing and creating!",
                    fg='yellow',
                    bg='black')
        classes, class_to_idx = _find_classes(root)
        imgs = _make_dataset(root, class_to_idx)
        _create_filelist(classes, imgs, root)
        fileset,_ = zip(*imgs)
        # fileset = _traverse_files(root, [])
        # _create_filelist(root, fileset)
    dataset_size = len(fileset)
    if typer == 'JPEG':
        fileset_check_method = _fileset_check
    else:
        fileset_check_method = _fileset_check_wav
    if process_num == 1:
        fileset_check_method(ctx.kspeedcli, objname, fileset, 0, dataset_size,
                             process_num, device)
        return

    proc = []
    queue = mp.Queue()
    imgs_per_process = int(dataset_size / pnum)
    for i in range(pnum - 1):
        p = mp.Process(target=_fileset_check,
                       args=(ctx.kspeedcli, objname, fileset,
                             i * imgs_per_process, (i + 1) * imgs_per_process,
                             process_num, device, queue))
        p.start()
        proc.append(p)
    p = mp.Process(target=_fileset_check,
                   args=(ctx.kspeedcli, objname, fileset,
                         (pnum - 1) * imgs_per_process, dataset_size,
                         process_num, device, queue))
    p.start()
    proc.append(p)

    for i in progressbar.progressbar(range(dataset_size),
                                     prefix='Check processing:'):
        queue.get()

    for i in proc:
        i.join()


# def _traverse_files(root, all_files):
#     # 首先遍历当前目录所有文件及文件夹
#     file_list = os.listdir(root)
#     # 准备循环判断每个元素是否是文件夹还是文件，是文件的话，把名称传入list，是文件夹的话，递归
#     for file in file_list:
#         if file[0] == '.':
#             continue
#         # 利用os.path.join()方法取得路径全名，并存入cur_path变量，否则每次只能遍历一层目录
#         cur_path = os.path.join(root, file)
#         # 判断是否是文件夹
#         if os.path.isdir(cur_path):
#             _traverse_files(cur_path, all_files)
#         else:
#             all_files.append(cur_path)
#     return all_files


# def _create_filelist(root, fileset):
#     f1 = open(os.path.join(root, '.fileset'), 'w')
#     for i in range(len(fileset)):
#         f1.write(fileset[i] + "\n")
#     f1.close()


def _read_filelist(root):
    # try:
    f1 = open(os.path.join(root, '.imgs'), 'r')
    lines = f1.readlines()
    fileset = []
    i = 0
    while i < len(lines):
        path=lines[i][:-1]
        i = i+3
        fileset.append(path)
    f1.close()
    return fileset
    # except Exception as e:
    #     raise e


def _fileset_check(kspeedcli,
                   dest,
                   fileset,
                   start,
                   end,
                   process_num,
                   device,
                   queue=None):
    if process_num == 1:
        bar = progressbar.ProgressBar(prefix='get processing:',
                                      max_value=(end - start)).start()
    dataset = kspeedcli.loadFileSet(dest)
    for i in range(start, end):
        path = fileset[i]
        f = open(path, 'rb')
        src_data = f.read()
        f.close()
        data_out = dataset.getBIN(path)
        if data_out is not None:
            if len(src_data) != len(data_out):
                click.secho(path + " check failed!!!!! length not same!",
                            fg='red',
                            bg='black')

            if data_out != src_data:
                click.secho(path + " check failed!!!!! pixel not same",
                            fg='red',
                            bg='black')

        if process_num == 1:
            bar.update(i)
        else:
            queue.put(i)


def _fileset_check_wav(kspeedcli,
                       dest,
                       fileset,
                       start,
                       end,
                       process_num,
                       device,
                       queue=None):
    if process_num == 1:
        bar = progressbar.ProgressBar(prefix='get processing:',
                                      max_value=(end - start)).start()
    dataset = kspeedcli.loadFileSet(dest, device)
    for i in range(start, end):
        path = fileset[i]
        src_data = np.fromfile(path, dtype='int16', count=-1, offset=44)
        data_out, _ = dataset.getWav(path)
        if data_out is not None:
            if len(src_data) != len(data_out):
                click.secho(path + " check failed!!!!! length not same!" +
                            str(len(src_data)) + "vs." + str(len(data_out)),
                            fg='red',
                            bg='black')
            if device == 'CPU':
                npt.assert_array_equal(src_data, data_out)
            else:
                import torch
                src_tensor = torch.from_numpy(src_data).cuda()
                torch.cuda.synchronize()
                data_tensor = torch.as_tensor(data_out,
                                              device=torch.device('cuda'))
                # print(f'data_tensor.shape : {data_tensor.shape}, data_ptr : {data_tensor.data_ptr()}, device : {data_tensor.device}')
                print(data_tensor.__cuda_array_interface__)
                torch.equal(data_tensor, src_tensor)

        if process_num == 1:
            bar.update(i)
        else:
            queue.put(i)


def _parse_root(root):
    if root[-1] == '/':
        root = root[:-1]
    if root[0] != '/':
        root = os.path.abspath(os.path.join(os.getcwd(), root))
    click.secho(f'root = {root}', fg='green', bg='black')
    return root

def _find_classes(directory: str):
    classes = sorted(entry.name for entry in os.scandir(directory)
                     if entry.is_dir())
    if not classes:
        raise FileNotFoundError(
            f"Couldn't find any class folder in {directory}.")

    class_to_idx = {cls_name: i for i, cls_name in enumerate(classes)}
    return classes, class_to_idx


def _make_dataset(directory: str, class_to_idx):
    directory = os.path.expanduser(directory)
    instances = []
    available_classes = set()
    for target_class in sorted(class_to_idx.keys()):
        class_index = class_to_idx[target_class]
        target_dir = os.path.join(directory, target_class)
        if not os.path.isdir(target_dir):
            continue
        for root, _, fnames in sorted(os.walk(target_dir, followlinks=True)):
            for fname in sorted(fnames):
                path = os.path.join(root, fname)
                item = path, class_index
                instances.append(item)

                if target_class not in available_classes:
                    available_classes.add(target_class)

    return instances

def _create_filelist(classes, imgs, root):
    root = _parse_root(root)
    f1 = open(root + '/.imgs', "w+")
    f2 = open(root + '/.classes', "w+")
    for i in range(len(classes)):
        s = classes[i]
        f2.write(s + "\n")
    for i in range(len(imgs)):
        f1.write(imgs[i][0] + "\n")
        f1.write(str(imgs[i][1]) + "\n")
        info = os.stat(imgs[i][0])
        f1.write(str(info.st_size) + "\n")
    f1.close()
    f2.close()
