import os
import sys
import progressbar
import multiprocessing as mp
import time
import warnings

from kspeed.kspeedcmd.kspeedcmd import pass_config
import click
import numpy.testing as npt
import numpy as np


@click.command("preload", short_help="Preload dataset to KSpeed.")
@click.option('-r',
              '--root',
              required=True,
              type=click.Path(exists=True),
              help="Specify the root path of dataset to preload to KSpeed.")
@click.option('-o',
              '--objname',
              type=str,
              help="Specify the object name in KSpeed.")
@click.option('-m',
              '--mode',
              default='w',
              type=click.Choice(['w', 'a']),
              help="Specify the mode to preload.")
@pass_config
def preload(ctx, root, objname, mode):
    """Preload dataset to KSpeed."""
    if root == "":
        click.secho("bad root path of dataset!", fg='red', bg='black')
        return
    root = _parse_path(root)
    ctx.kspeedcli.preload(root, objname, mode)


@click.command("check", short_help="check the dataset preloaded to KSpeed.")
@click.option('-r',
              '--root',
              required=True,
              help="Specify the root path or path list of dataset to check.")
@click.option('-o',
              '--objname',
              required=True,
              help="Specify the object name to get and check from KSpeed.")
@click.option('-m',
              '--mode',
              type=click.Choice(['get', 'getbatch']),
              default='get',
              help="Specify the mode of get from KSpeed.")
@click.option('-b',
              '--batch_size',
              type=int,
              default=1,
              help="Specify the batch size of getbatch from KSpeed.")
@click.option('-n',
              "process_num",
              default=10,
              type=int,
              help="Specify the process num used to get and check.")
@click.option('-d',
              '--device',
              type=click.Choice(['CPU', 'GPU']),
              default='CPU',
              help="Specify the device to get and check from KSpeed.")
@click.option(
    '-t',
    '--typer',
    type=click.Choice(['NUMPY', 'JPEG', 'BIN', 'WAV']),
    default='JPEG',
    help="Specify the get api type of the object to get and check from KSpeed."
)
@pass_config
def check(ctx, root, objname, mode, batch_size, process_num, device, typer):
    """check the fileset preloaded to KSpeed."""
    if root == "":
        click.secho("bad root data path!", fg='red', bg='black')
        return
    if not os.path.isdir(root):
        _npy_check(ctx, root, objname, mode, batch_size, 1, device,
                   process_num)
        return

    root = _parse_path(root)
    pnum = process_num
    try:
        fileset = _read_filelist(root)
        click.secho("fileset found!", fg='green', bg='black')
    except IOError:
        click.secho("No fileset found, traversing and creating!",
                    fg='yellow',
                    bg='black')
        classes, class_to_idx = _find_classes(root)
        imgs = _make_dataset(root, class_to_idx)
        _create_filelist(classes, imgs, root)
        fileset,_ = zip(*imgs)
        # fileset = _traverse_files(root, [])
        # _create_filelist(root, fileset)
    dataset_size = len(fileset)
    if typer == 'JPEG':
        fileset_check_method = _fileset_check
    else:
        fileset_check_method = _fileset_check_wav
    if process_num == 1:
        fileset_check_method(ctx.kspeedcli, objname, fileset, 0, dataset_size,
                             process_num, device)
        return

    proc = []
    queue = mp.Queue()
    imgs_per_process = int(dataset_size / pnum)
    for i in range(pnum - 1):
        p = mp.Process(target=_fileset_check,
                       args=(ctx.kspeedcli, objname, fileset,
                             i * imgs_per_process, (i + 1) * imgs_per_process,
                             process_num, device, queue))
        p.start()
        proc.append(p)
    p = mp.Process(target=_fileset_check,
                   args=(ctx.kspeedcli, objname, fileset,
                         (pnum - 1) * imgs_per_process, dataset_size,
                         process_num, device, queue))
    p.start()
    proc.append(p)

    for i in progressbar.progressbar(range(dataset_size),
                                     prefix='Check processing:'):
        queue.get()

    for i in proc:
        i.join()


def _traverse_files(root, all_files):
    # 首先遍历当前目录所有文件及文件夹
    file_list = os.listdir(root)
    # 准备循环判断每个元素是否是文件夹还是文件，是文件的话，把名称传入list，是文件夹的话，递归
    for file in file_list:
        if file[0] == '.':
            continue
        # 利用os.path.join()方法取得路径全名，并存入cur_path变量，否则每次只能遍历一层目录
        cur_path = os.path.join(root, file)
        # 判断是否是文件夹
        if os.path.isdir(cur_path):
            _traverse_files(cur_path, all_files)
        else:
            all_files.append(cur_path)
    return all_files


# def _create_filelist(root, fileset):
#     f1 = open(os.path.join(root, '.fileset'), 'w')
#     for i in range(len(fileset)):
#         f1.write(fileset[i] + "\n")
#     f1.close()

def _create_filelist(classes, imgs, root):
    root = _parse_path(root)
    f1 = open(root + '/.imgs', "w+")
    f2 = open(root + '/.classes', "w+")
    for i in range(len(classes)):
        s = classes[i]
        f2.write(s + "\n")
    for i in range(len(imgs)):
        f1.write(imgs[i][0] + "\n")
        f1.write(str(imgs[i][1]) + "\n")
        info = os.stat(imgs[i][0])
        f1.write(str(info.st_size) + "\n")
    f1.close()
    f2.close()


def _read_filelist(root):
    try:
        f1 = open(os.path.join(root, '.imgs'), 'r')
        lines = f1.readlines()
        fileset = []
        i = 0
        while i < len(lines):
            path=lines[i][:-1]
            i = i+3
            fileset.append(path)
        f1.close()
        return fileset
    except Exception as e:
        raise e


def _fileset_check(kspeedcli,
                   dest,
                   fileset,
                   start,
                   end,
                   process_num,
                   device,
                   queue=None):
    if process_num == 1:
        bar = progressbar.ProgressBar(prefix='checking processing:',
                                      max_value=(end - start)).start()
    dataset = kspeedcli.load(dest)
    for i in range(start, end):
        path = fileset[i]
        f = open(path, 'rb')
        src_data = f.read()
        f.close()
        data_out = dataset.getBIN(path)
        if data_out is not None:
            if len(src_data) != len(data_out):
                click.secho(path + " check failed!!!!! length not same!",
                            fg='red',
                            bg='black')

            if data_out != src_data:
                click.secho(path + " check failed!!!!! pixel not same",
                            fg='red',
                            bg='black')

        if process_num == 1:
            bar.update(i)
        else:
            queue.put(i)


def _fileset_check_wav(kspeedcli,
                       dest,
                       fileset,
                       start,
                       end,
                       process_num,
                       device,
                       queue=None):
    if process_num == 1:
        bar = progressbar.ProgressBar(prefix='get processing:',
                                      max_value=(end - start)).start()
    dataset = kspeedcli.load(dest, device)
    for i in range(start, end):
        path = fileset[i]
        src_data = np.fromfile(path, dtype='int16', count=-1, offset=44)
        data_out, _ = dataset.getWav(path)
        if data_out is not None:
            if len(src_data) != len(data_out):
                click.secho(path + " check failed!!!!! length not same!" +
                            str(len(src_data)) + "vs." + str(len(data_out)),
                            fg='red',
                            bg='black')
            if device == 'CPU':
                npt.assert_array_equal(src_data, data_out)
            else:
                import torch
                src_tensor = torch.from_numpy(src_data).cuda()
                torch.cuda.synchronize()
                data_tensor = torch.as_tensor(data_out,
                                              device=torch.device('cuda'))
                # print(f'data_tensor.shape : {data_tensor.shape}, data_ptr : {data_tensor.data_ptr()}, device : {data_tensor.device}')
                print(data_tensor.__cuda_array_interface__)
                torch.equal(data_tensor, src_tensor)

        if process_num == 1:
            bar.update(i)
        else:
            queue.put(i)


def _parse_path(root):
    if root[-1] == '/':
        root = root[:-1]
    if root[0] != '/':
        root = os.path.abspath(os.path.join(os.getcwd(), root))
    click.secho(f'root = {root}', fg='green', bg='black')
    return root


def _npy_check(ctx, filelist, objname, mode, batch_size, prefetch_num, device,
               process_num):
    """check the fileset preloaded to KSpeed."""
    # mp.set_start_method('spawn')
    if device == 'GPU' and process_num > 1:
        process_num = 1
        warnings.warn('process_num option could only be 1 in GPU mode.'
                      'set process_num == 1 forcely to continue GPU mode test.')
    filelist = filelist.split(',')
    for f in filelist:
        if not os.path.exists(f):
            raise ValueError(f"{f} is not accessible.!")

    make_dataset = mp.Manager().dict()
    idrange_pre = 0
    for i in range(len(filelist)):
        data_npy = np.load(filelist[i], mmap_mode='r')
        data_npy = data_npy.reshape(-1, data_npy.shape[-1])
        id_range = len(data_npy)
        for j in range(id_range):
            make_dataset[idrange_pre + j] = [filelist[i], j]
        idrange_pre += id_range
    dataset_size = len(make_dataset)
    if mode == 'get':
        func = _npy_get_check
        assert batch_size == 1, "batch_size must be equal to 1 in get mode"
    else:
        func = _npy_getbatch_check
    if process_num == 1:
        func(0, ctx.kspeedcli, make_dataset, 0, dataset_size, objname,
             batch_size, prefetch_num, device)
        return

    idx_per_process = int(dataset_size / process_num)
    queue = mp.Queue()
    proc = []
    kspeedcli = ctx.kspeedcli
    for i in range(process_num - 1):
        p = mp.Process(target=func,
                       args=(i, kspeedcli, make_dataset, i * idx_per_process,
                             (i + 1) * idx_per_process, objname, batch_size,
                             prefetch_num, device, queue))
        p.start()
        proc.append(p)
    p = mp.Process(target=func,
                   args=(process_num - 1, kspeedcli, make_dataset,
                         (process_num - 1) * idx_per_process, dataset_size,
                         objname, batch_size, prefetch_num, device, queue))
    p.start()
    proc.append(p)

    for i in progressbar.progressbar(range(dataset_size),
                                     prefix='Checking processing:'):
        queue.get()

    for i in proc:
        i.join()


def _npy_get_check(pid,
                   kspeedcli,
                   make_dataset,
                   start,
                   end,
                   target,
                   batch_size,
                   prefetch_num,
                   device,
                   queue=None):
    if device == 'GPU':
        import torch
        cuda_device_cnt = torch.cuda.device_count()
        if cuda_device_cnt == 0:
            click.secho("No cuda device detect, use CPU", fg='red', bg='black')
            device == 'CPU'
        else:
            torch.cuda.set_device(pid % cuda_device_cnt)
    dataset = kspeedcli.load(target, device, batch_size, prefetch_num)
    i = start
    bar = None
    idx_for_bar = 0
    if queue is None:
        # widgets = [
        #     progressbar.Percentage(),
        #     progressbar.Bar(), ' [',
        #     progressbar.Timer(), '] '
        # ]
        bar = progressbar.ProgressBar(  #widgets=widgets,
            prefix='Check processing:',
            max_value=end - start).start()

    while i < end:
        left_num = end - i
        current_file = make_dataset[i][0]
        start_idx_in_current_file = make_dataset[i][1]
        data_npy = np.load(current_file, mmap_mode='r')
        data_npy = data_npy.reshape(-1, data_npy.shape[-1])
        id_range = len(data_npy) - start_idx_in_current_file
        id_range_in_process = min(left_num, id_range)
        end_idx_in_current_file = start_idx_in_current_file + id_range_in_process
        for j in range(start_idx_in_current_file, end_idx_in_current_file):
            try:
                data = dataset.get(i)
                # print(data.__cuda_array_interface__)
                if device == 'CPU':
                    npt.assert_array_equal(data, data_npy[j])
                else:
                    import torch
                    src_tensor = torch.from_numpy(data_npy[j]).cuda()
                    torch.cuda.synchronize()
                    data_tensor = torch.as_tensor(data,
                                                  device=torch.device('cuda'))
                    # print(f'data_tensor.shape : {data_tensor.shape}, data_ptr : {data_tensor.data_ptr()}, device : {data_tensor.device}')
                    # print(data_tensor.__cuda_array_interface__)
                    torch.equal(data_tensor, src_tensor)
            except Exception as e:
                print(e)
                print("ERROR: kspeed load and check failed, idx=", i)
                sys.exit(-1)
            i = i + 1
            if queue is None:
                idx_for_bar = idx_for_bar + 1
                bar.update(idx_for_bar)
            else:
                queue.put(idx_for_bar)


def _npy_getbatch_check(pid,
                        kspeedcli,
                        make_dataset,
                        start,
                        end,
                        target,
                        batch_size,
                        prefetch_num,
                        device,
                        queue=None):
    if device == 'GPU':
        import torch
        cuda_device_cnt = torch.cuda.device_count()
        if cuda_device_cnt == 0:
            click.secho("No cuda device detect, use CPU", fg='red', bg='black')
            device == 'CPU'
        else:
            torch.cuda.set_device(pid % cuda_device_cnt)
    dataset = kspeedcli.load(target, device, batch_size, prefetch_num)
    i = start
    bar = None
    idx_for_bar = 0
    if queue is None:
        # widgets = [
        #     progressbar.Percentage(),
        #     progressbar.Bar(), ' [',
        #     progressbar.Timer(), '] '
        # ]
        bar = progressbar.ProgressBar(  #widgets=widgets,
            prefix='Check processing:',
            max_value=end - start).start()

    while i < end:
        left_num = end - i
        current_file = make_dataset[i][0]
        start_idx_in_current_file = make_dataset[i][1]
        data_npy = np.load(current_file, mmap_mode='r')
        data_npy = data_npy.reshape(-1, data_npy.shape[-1])
        id_range = len(data_npy) - start_idx_in_current_file
        id_range_in_process = min(left_num, id_range)
        end_idx_in_current_file = start_idx_in_current_file + id_range_in_process
        for j in range(start_idx_in_current_file, end_idx_in_current_file,
                       batch_size):
            if j + batch_size > end_idx_in_current_file:
                batch_size_t = end_idx_in_current_file - j
            else:
                batch_size_t = batch_size
            try:
                data = dataset.getbatch(np.arange(i, i + batch_size_t),
                                        batch_size_t)
                # print(data.__cuda_array_interface__)
                if device == 'CPU':
                    npt.assert_array_equal(data, data_npy[j:j + batch_size_t])
                else:
                    import torch
                    src_tensor = torch.from_numpy(
                        data_npy[j:j + batch_size_t]).cuda()
                    torch.cuda.synchronize()
                    data_tensor = torch.as_tensor(data,
                                                  device=torch.device('cuda'))
                    # print(f'data_tensor.shape : {data_tensor.shape}, data_ptr : {data_tensor.data_ptr()}, device : {data_tensor.device}')
                    # print(data_tensor.__cuda_array_interface__)
                    torch.equal(data_tensor, src_tensor)
            except Exception as e:
                click.secho(e, fg='red', bg='black')
                click.secho("ERROR: kspeed load and check failed, idx=",
                            i,
                            fg='red',
                            bg='black')
                sys.exit(-1)
            if queue is None:
                idx_for_bar = idx_for_bar + batch_size_t
                bar.update(idx_for_bar)
            else:
                for h in range(i, i + batch_size_t):
                    queue.put(h)
            i = i + batch_size_t


@click.command("get", short_help="get the dataset sample from KSpeed.")
@click.option('-r',
              '--root',
              required=True,
              help="Specify the root path or path list of dataset to get.")
@click.option('-o',
              '--objname',
              required=True,
              help="Specify the object name to get from KSpeed.")
@click.option('-n',
              "process_num",
              default=10,
              type=int,
              help="Specify the process num used to get.")
@click.option('-d',
              '--device',
              type=click.Choice(['CPU', 'GPU']),
              default='CPU',
              help="Specify the device to get from KSpeed.")
@click.option('-t',
              '--typer',
              type=click.Choice(['NUMPY', 'PIL', 'BIN', 'WAV']),
              default='PIL',
              help="Specify the get api type of the object to get from KSpeed."
              )
@pass_config
def get(ctx, root, objname, process_num, device, typer):
    """get the dataset that preloaded to KSpeed."""
    if device == 'GPU' and process_num > 1:
        raise ValueError('process_num option could only be 1 in GPU mode.'
                         'let process_num == 1 to enable GPU mode.')
    pnum = process_num
    fileset = None
    dataset_size = 0
    if not os.path.isdir(root):
        root = _parse_path(root)
        try:
            fileset = _read_filelist(root)
            click.secho("fileset found!", fg='green', bg='black')
        except IOError:
            click.secho("No fileset found, traversing and creating!",
                        fg='yellow',
                        bg='black')
            classes, class_to_idx = _find_classes(root)
            imgs = _make_dataset(root, class_to_idx)
            _create_filelist(classes, imgs, root)
            fileset,_ = zip(*imgs)
            # fileset = _traverse_files(root, [])
            # _create_filelist(root, fileset)
        dataset_size = len(fileset)

    if process_num == 1:
        _get(ctx.kspeedcli, objname, fileset, 0, process_num, device, typer)
        return

    proc = []
    queue = mp.Queue()
    dataset_size = mp.Value('i', 0)
    lock = mp.Lock()
    for i in range(pnum):
        p = mp.Process(target=_get,
                       args=(ctx.kspeedcli, objname, fileset, i, process_num,
                             device, typer, queue, dataset_size, lock))
        p.start()
        proc.append(p)

    while dataset_size.value == 0:
        if dataset_size.value > 0:
            break
        time.sleep(0.1)
    for i in progressbar.progressbar(range(dataset_size.value),
                                     prefix='get processing:'):
        queue.get()

    for i in proc:
        i.join()


def _get(kspeedcli,
         objname,
         fileset,
         pid,
         process_num,
         device,
         typer,
         queue=None,
         dataset_size=None,
         lock=None):
    dataset = kspeedcli.load(objname, device)
    dataset_size_i = dataset.size()
    if fileset is not None:
        if dataset_size_i != len(fileset):  # FIXME
            print(
                f'fileset sizeis not equal the actual size {dataset_size_i} !=  {len(fileset)}'
            )
            dataset_size_i = len(fileset)

    if lock is not None:
        lock.acquire()
        dataset_size.value = dataset_size_i
        lock.release()

    idx_per_process = int(dataset_size_i / process_num)
    start = pid * idx_per_process
    end = (pid + 1) * idx_per_process
    if pid == process_num - 1:
        end = dataset_size_i

    bar = None
    if queue is None:
        bar = progressbar.ProgressBar(  #widgets=widgets,
            prefix='get processing:',
            max_value=end - start).start()

    for i in range(start, end):
        if fileset is None:
            data = dataset.get(i)
        else:
            if typer == 'JPEG':
                data = dataset.getPIL(fileset[i])
            elif typer == 'BIN':
                data = dataset.getBIN(fileset[i])
            elif typer == 'NUMPY':
                data = dataset.get(fileset[i])
            else:  #typer == 'WAV'
                data = dataset.getWav(fileset[i])

        if queue is None:
            bar.update(i)
        else:
            queue.put(i)


def _find_classes(directory: str):
    classes = sorted(entry.name for entry in os.scandir(directory)
                     if entry.is_dir())
    if not classes:
        raise FileNotFoundError(
            f"Couldn't find any class folder in {directory}.")

    class_to_idx = {cls_name: i for i, cls_name in enumerate(classes)}
    return classes, class_to_idx

def _make_dataset(directory: str, class_to_idx):
    directory = os.path.expanduser(directory)
    instances = []
    # available_classes = set()
    for target_class in sorted(class_to_idx.keys()):
        class_index = class_to_idx[target_class]
        target_dir = os.path.join(directory, target_class)
        if not os.path.isdir(target_dir):
            continue
        for root, _, fnames in sorted(os.walk(target_dir, followlinks=True)):
            for fname in sorted(fnames):
                path = os.path.join(root, fname)
                item = path, class_index
                instances.append(item)

                # if target_class not in available_classes:
                #     available_classes.add(target_class)

    return instances

