import os
import configparser
from kspeed.kspeedcmd.kspeedcmd import pass_config
import re
import click

CONFIGFILE = "%s/.kspeedcredentials" % os.path.expanduser('~')
CONFIGSECTION = 'KSpeedCredentials'
CONTEXT_SETTINGS = dict(auto_envvar_prefix="COMPLEX")
OPTION_DICT = {
    "-u/--username": "Specify the username to login.",
    "-w/--password": "Specify the password to login.",
    "-i/--iplist": "specify the kspeed server cluster iplist.",
    "-o/--objname": "Specify the object name  in KSpeed.",
}


def abort_if_false(ctx, param, value):
    if not value:
        ctx.abort()


@click.command(
    "config",
    short_help="Create configuration file to resource access credentials.")
@click.option("-u",
              "--username",
              required=True,
              help="Specify the username to login.")
@click.option("-w",
              "--password",
              required=True,
              help="Specify the password to login.")
@click.option("-i",
              "--iplist",
              help="specify the kspeed server cluster iplist.")
def config(username, password, iplist):
    """Create configuration file to resource access credentials."""
    try:
        config = configparser.RawConfigParser()
        config.add_section(CONFIGSECTION)
        config.set(CONFIGSECTION, 'username', username)
        config.set(CONFIGSECTION, 'password', password)
        if iplist is not None:
            config.set(CONFIGSECTION, 'iplist', iplist)
        cfgfile = open(CONFIGFILE, 'w')
        config.write(cfgfile)
        click.secho("Your configuration is saved into %s ." % CONFIGFILE,
                    fg='green',
                    bg='black')
        cfgfile.close()
        import stat
        os.chmod(CONFIGFILE, stat.S_IREAD | stat.S_IWRITE)
    except Exception as e:
        click.secho("config Failed!", e, fg='red', bg='black')


@click.command("adduser", short_help="Add new user to KSpeed.")
@click.option("-u",
              "--username",
              required=True,
              help="Specify the username to login.")
@click.option("-w",
              "--password",
              required=True,
              help="Specify the password to login.")
@pass_config
def adduser(ctx, username, password):
    """Add new user to KSpeed."""
    ret = ctx.kspeedcli.adduser(username, password)
    if ret == 0:
        click.secho("User added successfully", fg='green', bg='black')
    else:
        click.secho("Failed to add user", fg='red', bg='black')


@click.command("deluser", short_help="Delete user from KSpeed.")
@click.option("-u",
              "--username",
              required=True,
              help="Specify the username to login.")
@pass_config
@click.option("-yes",
              is_flag=True,
              callback=abort_if_false,
              expose_value=False,
              prompt='Are you sure you want to drop the user?')
def deluser(ctx, username):
    """Delete user from KSpeed."""
    ret = ctx.kspeedcli.deluser(username)
    if ret == 0:
        click.secho("User deleted successfully", fg='green', bg='black')
    else:
        click.secho("Failed to delete user", fg='red', bg='black')


@click.command("passwd", short_help="Change password of user in KSpeed.")
@click.option("-w",
              "--password",
              required=True,
              help="Specify the username to login.")
@pass_config
def passwd(ctx, password):
    """Change password of user in KSpeed."""
    ret = ctx.kspeedcli.passwd(password)
    if ret == 0:
        click.secho("Password modified successfully", fg='green', bg='black')
    else:
        click.secho("Failed to modify password", fg='red', bg='black')


@click.command("rm", short_help="Remove a dataset from KSpeed.")
@click.option("-o",
              "--objname",
              required=True,
              help="Specify the object name to remove in KSpeed.")
@click.option("-yes",
              is_flag=True,
              callback=abort_if_false,
              expose_value=False,
              prompt='Are you sure you want to remove a dataset from KSpeed?')
@pass_config
def rm(ctx, objname):
    """Remove dataset from KSpeed."""
    objlist = re.split('\ |\,|\;', objname)
    for obj in objlist:
        ret = ctx.kspeedcli.rm(obj)
        if ret == 0:
            click.secho(f"{obj} deleted from kspeed successfully",
                        fg='green',
                        bg='black')
        else:
            click.secho(f"Failed to delete {obj} from kspeed",
                        fg='red',
                        bg='black')
