import os
import sys
import csv
import torch
import torch.distributed as dist
from torch.distributed import get_rank, get_world_size
from kspeed.kspeedcluster import kspeedclient
import multiprocessing as mp
from PIL import Image
import numpy as np


class KSpeedDataset(object):

    def __init__(self,
                 root,
                 object_name=None,
                 dl_workers=4,
                 device='CPU',
                 batchsize=1,
                 kspeed_iplist=None,
                 username='admin',
                 password='admin'):
        if root is None and object_name is None:
            raise ValueError("root and object_name are both None, exit!")
        if root is not None:
            root = _parse_root(root)
        self.root = root
        if object_name is None:
            self.object_name = (root.split('/'))[-1]
        else:
            self.object_name = object_name
        self.username = username
        self.password = password
        self.dl_workers = dl_workers
        self.device = device
        self.batchsize = batchsize
        self.auto_preload_unload = False
        if kspeed_iplist is None:
            self.kspeedcli = kspeedclient(None, self.username, self.password)
        else:
            with open(kspeed_iplist, 'r') as f:
                kspeed_iplist_str = ' '.join(f.read().split())
            self.kspeed_iplist_str = kspeed_iplist_str
            self.kspeedcli = kspeedclient(str(self.kspeed_iplist_str),
                                          self.username, self.password)
        self.dataset = None
        if torch.distributed.is_initialized():
            world_size = get_world_size()
            self.rank = get_rank()
        else:
            if os.getenv('WORLD_SIZE') is None:
                world_size = 1
            else:
                world_size = int(os.environ['WORLD_SIZE'])
            if os.getenv('RANK') is None:
                self.rank = 0
            else:
                self.rank = int(os.environ['RANK'])
        os.environ['KSPEED_WORLD_SIZE'] = str(world_size)
        os.environ['KSPEED_RANK'] = str(self.rank)
        cuda_device = torch.cuda.current_device()
        gpu_bus_id = self.kspeedcli.get_gpu_busid(cuda_device)
        os.environ['KSPEED_GPU_BUSID'] = str(gpu_bus_id)

        self.size = self.get_size()
        if self.size == 0:
            if torch.distributed.is_initialized():
                dist.barrier()
                print(
                    f"[{os.getpid()}] : get_size(={self.size}) barrier finished at rank {self.rank}"
                )

            self.auto_preload_unload = True
            if self.rank == 0:
                self.prepare_preload()
            if torch.distributed.is_initialized():
                dist.barrier()
                print(
                    f"[{os.getpid()}] : prepare_preload barrier finished at rank {self.rank}"
                )

            self.preload()
            if torch.distributed.is_initialized():
                dist.barrier()
                print(
                    f"[{os.getpid()}] : preload barrier finished at rank {self.rank}"
                )
        self.size = self.get_size()
        print(f"[{os.getpid()}] : get_size = {self.size} at rank {self.rank}")

    def get_size(self):
        if self.dl_workers == 0:
            return self._get_size()
        else:
            mp_size = mp.Value('i', 0)
            p = mp.Process(target=self._get_size, args=(mp_size, ))
            p.start()
            p.join()
            return mp_size.value

    def _get_size(self, size=None):
        self.dataset = self.kspeedcli.load(self.object_name, self.device,
                                           self.batchsize)
        if self.dataset is None:
            dataset_size = 0
        else:
            dataset_size = self.dataset.size()
        if size is None:
            return dataset_size
        else:
            size.value = dataset_size
            return size.value

    def prepare_preload(self):
        if self.dl_workers == 0:
            return self._prepare_preload()
        else:
            p = mp.Process(target=self._prepare_preload)
            p.start()
            p.join()

    def _prepare_preload(self):
        self.dataset = self.kspeedcli.prepare_preload(self.root,
                                                      self.object_name)

    def preload(self):
        if self.dl_workers == 0:
            self._preload()
        else:
            p = mp.Process(target=self._preload)
            p.start()
            p.join()

    def _preload(self):
        self.dataset = self.kspeedcli.preload(self.root, self.object_name)

    def load(self):
        self.dataset = self.kspeedcli.load(self.object_name, self.device,
                                           self.batchsize)
        # print(
        #     f"load {self.object_name} with size = {self.dataset.size()} to {self.device}"
        # )
        if self.dataset.size() == 0:
            raise RuntimeError(f'dataset {self.object_name} is empty, exit!')

    def __getitem__(self, idx):
        sample = None
        try:
            if self.dataset is None:
                self.load()

            if not isinstance(idx, int):
                idx_a = np.array(idx)
                sample = self.dataset.getbatch(idx_a)
            else:
                sample = self.dataset.get(idx)
        except Exception as e:
            print("ERROR: kspeed read failed : {}".format(e))
            sys.exit()

        if sample is None:
            print("ERROR: kspeed read miss, idx : {}".format(idx))

        return sample

    def __len__(self):
        return self.size

    def rm(self):
        if self.dl_workers == 0:
            self._rm()
        else:
            p = mp.Process(target=self._rm)
            p.start()
            p.join()

    def _rm(self):
        ret = self.kspeedcli.rm(self.object_name)
        if ret == 0:
            print(f"Auto-unloading {self.object_name} successfully")
        else:
            print(f"Auto-unloading {self.object_name} Failed")

    def __del__(self):
        if self.auto_preload_unload and self.rank == 0:
            self.rm()


class KSpeedFolder(KSpeedDataset):

    def __init__(self,
                 root,
                 object_name=None,
                 dl_workers=4,
                 kspeed_iplist=None,
                 username='admin',
                 password='admin'):
        try:
            classes, class_to_idx, files = read_filelist(root)
            if files is None:
                print("files is None, recreating filelist")
                os.remove(root+'/.imgs')
                os.remove(root+'/.classes')
                classes, class_to_idx = find_classes(root)
                files = make_dataset(root, class_to_idx)
                create_filelist(classes, files, root)
            # print("filelist found!")
        except Exception:
            print("No filelist found, creating!")
            classes, class_to_idx = find_classes(root)
            files = make_dataset(root, class_to_idx)
            create_filelist(classes, files, root)
        if len(files) == 0:
            raise RuntimeError("Found 0 sample in subfolders of: " + root +
                               "\n")
        #先产生辅助文件，再调用基类初始化函数，保证所有节点都有辅助文件以后再preload
        super(KSpeedFolder,
              self).__init__(root, object_name, dl_workers, 'CPU', 1,
                             kspeed_iplist, username, password)
        self.files = files
        self.classes = classes
        self.class_to_idx = class_to_idx

    def __getitem__(self, index):
        """
        index (int): Index
        Returns:tuple: (file_b, target) where target is class_index of the target class.
        """
        if self.dataset is None:
            self.load()

        if not isinstance(index, int):
            return []
        else:
            path, target = self.files[index]  #获取图像（路径，标签）
            file_b = self.dataset.getBIN(path)  #数据读取
            if file_b is None:
                print("KSpeed Read dataout is None")
            return file_b, target

    def __len__(self):
        return len(self.files)


class KSpeedImageFolder(KSpeedFolder):

    def __init__(self,
                 root,
                 object_name=None,
                 dl_workers=4,
                 kspeed_iplist=None,
                 username='admin',
                 password='admin',
                 transform=None,
                 target_transform=None):
        super(KSpeedImageFolder,
              self).__init__(root, object_name, dl_workers, kspeed_iplist,
                             username, password)
        self.transform = transform
        self.target_transform = target_transform

    def __getitem__(self, index):
        """
        index (int): Index
        Returns:tuple: (image, target) where target is class_index of the target class.
        """
        if self.dataset is None:
            self.load()

        if not isinstance(index, int):
            return []
        else:
            path, target = self.files[index]  #获取图像（路径，标签）
            img = self.dataset.getPIL(path)  #数据读取
            if img is not None:
                if img.mode != 'RGB':
                    img = img.convert('RGB')
            else:
                img = Image.new("RGB", (244, 244))
                print("KSpeed Read dataout is None")
            if self.transform is not None:
                img = self.transform(img)  #数据、标签转换成tensor
            if self.target_transform is not None:
                target = self.target_transform(target)
            return img, target


def find_classes(directory: str):
    classes = sorted(entry.name for entry in os.scandir(directory)
                     if entry.is_dir())
    if not classes:
        raise FileNotFoundError(
            f"Couldn't find any class folder in {directory}.")

    class_to_idx = {cls_name: i for i, cls_name in enumerate(classes)}
    return classes, class_to_idx


def make_dataset(directory: str, class_to_idx):
    directory = os.path.expanduser(directory)
    instances = []
    available_classes = set()
    for target_class in sorted(class_to_idx.keys()):
        class_index = class_to_idx[target_class]
        target_dir = os.path.join(directory, target_class)
        if not os.path.isdir(target_dir):
            continue
        for root, _, fnames in sorted(os.walk(target_dir, followlinks=True)):
            for fname in sorted(fnames):
                path = os.path.join(root, fname)
                item = path, class_index
                instances.append(item)

                if target_class not in available_classes:
                    available_classes.add(target_class)

    return instances


def read_filelist(root):
    root = _parse_root(root)
    f1 = open(root + '/.imgs', 'r')
    f2 = open(root + '/.classes', 'r')
    lines = f1.readlines()
    if len(lines) % 3 != 0:
        raise RuntimeError("Broken file .imgs")
    if lines[0][:len(root)] != root:
        return None, None, None
    imgs = []
    i = 0
    while i < len(lines):
        path=lines[i][:-1]
        label=lines[i+1][:-1]
        i = i+3
        imgs.append((path, int(label)))

    lines = f2.readlines()
    classes = []
    class_to_idx = {}
    i = 0
    for line in lines:
        classes.append(line[:-1])
        class_to_idx[line[:-1]] = i
        i = i + 1
    f1.close()
    f2.close()
    return classes, class_to_idx, imgs


def create_filelist(classes, imgs, root):
    root = _parse_root(root)
    f1 = open(root + '/.imgs', "w+")
    f2 = open(root + '/.classes', "w+")
    for i in range(len(classes)):
        s = classes[i]
        f2.write(s + "\n")
    for i in range(len(imgs)):
        f1.write(imgs[i][0] + "\n")
        f1.write(str(imgs[i][1]) + "\n")
        info = os.stat(imgs[i][0])
        f1.write(str(info.st_size) + "\n")
    f1.close()
    f2.close()


def _parse_root(root):
    if root[-1] == '/':
        root = root[:-1]
    if root[0] != '/':
        root = os.path.abspath(os.path.join(os.getcwd(), root))
    # print(f'root = {root}')
    return root


class KSpeedOpenImageFolder(KSpeedDataset):

    def __init__(self,
                 kspeed_iplist,
                 username=None,
                 password=None,
                 root=None,
                 object_name=None,
                 transform=None,
                 target_transform=None):
        if root is None and object_name is None:
            raise ValueError("root and object_name are both None, exit!")
        if root is not None and object_name is not None:
            raise ValueError("root and object_name are both not None, exit!")
        super(KSpeedOpenImageFolder, self).__init__(kspeed_iplist, username,
                                                    password)
        self.root = root
        self.object_name = object_name
        self.transform = transform
        self.target_transform = target_transform
        imgs_dict = {}
        class_to_idx = {}
        with open(root + '/train-annotations-human-imagelabels-boxable.csv',
                  'r') as csvfile:
            f = csv.reader(csvfile, delimiter=',')
            for item in f:
                [id, _, label, confidence] = item
                if id == "ImageID":
                    continue
                if confidence == '1':
                    if label not in class_to_idx:
                        class_to_idx[label] = len(class_to_idx)
                    if id not in imgs_dict:
                        imgs_dict[id] = label
                    # else:
                    #     imgs_dict[id].append(label)
        imgs = sorted(imgs_dict.keys())
        if len(imgs) == 0:
            raise RuntimeError("Found 0 images in subfolders of: " + root +
                               "\n")
        self.imgs = imgs
        self.class_to_idx = class_to_idx
        self.imgs_dict = imgs_dict

    def __getitem__(self, index):
        if self.dataset is None:
            self.imageset_load(self.root, self.object_name)

        path = self.imgs[index]  #获取图像（路径，标签）
        t = self.imgs_dict[path]
        target = self.class_to_idx[t]
        img = self.dataset.get(self.object_name,
                               self.root + '/train/' + path + '.jpg')  #数据读取

        if img.mode != 'RGB':
            img = img.convert('RGB')
        if self.transform is not None:
            img = self.transform(img)  #数据、标签转换成tensor
        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self):
        return len(self.imgs)