import time
import os
import sys
import argparse
import numpy as np
import numpy.testing as npt
import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from torch.distributed import get_rank
from torch.utils.data import BatchSampler, RandomSampler, SequentialSampler

from kspeed.utils.data.kspeeddataset import KSpeedDataset, KSpeedFolder

import warnings

warnings.filterwarnings('ignore')


def setup_training(local_rank, ngpus):
    # initialize the process group
    # torch.cuda.set_device(local_rank)

    if os.getenv('MASTER_ADDR') is None:
        os.environ['MASTER_ADDR'] = '127.0.0.1'
    if os.getenv('MASTER_PORT') is None:
        os.environ['MASTER_PORT'] = '9785'
    if os.getenv('RANK') is None:
        os.environ['RANK'] = '0'
    if os.getenv('WORLD_SIZE') is None:
        os.environ['WORLD_SIZE'] = '1'
    master_ip = os.environ['MASTER_ADDR']
    master_port = os.environ['MASTER_PORT']
    worker_size = int(os.environ['WORLD_SIZE'])
    worker_rank = int(os.environ['RANK'])
    ngpus_per_worker = ngpus  #int(os.environ['KUBERNETES_CONTAINER_RESOURCE_GPU'])
    world_size = worker_size * ngpus_per_worker
    world_rank = worker_rank * ngpus_per_worker + local_rank
    init_method = 'tcp://' + master_ip + ':' + master_port
    print(
        f"[{os.getpid()}] Initializing process group with: master_ip:{master_ip}, master_port:{master_port}, worker_size:{worker_size},worker_rank:{worker_rank}, ngpu_per_worker:{ngpus_per_worker}, init_method:{init_method}, world_size:{world_size}, world_rank:{world_rank}, local_rank:{local_rank}"
    )
    dist.init_process_group(backend='gloo',
                            world_size=world_size,
                            rank=world_rank)
    print(
        f"[{os.getpid()}] : world_rank{world_rank} start all_reduce for init_process_group"
    )
    tensor = torch.ones(2, 16) * get_rank()
    dist.all_reduce(tensor, op=dist.ReduceOp.SUM, async_op=False)
    dist.barrier()
    print(
        f"[{os.getpid()}] : world_rank{world_rank} finished all_reduce for init_process_group"
    )


#----------------------------training part ----------------------------
def data_loading(dl, args):
    for e in range(args.epoch_num):
        for i, data in enumerate(dl):
            if len(data) == 2:
                batch, target = data
            else:
                batch = data
            if i == 0:
                time_s1 = int(round(time.time() * 1000000))
            else:
                cost_time0 = int(round(time.time() * 1000000)) - time_s1
                time_s1 = int(round(time.time() * 1000000))
                if i % 1e2 == 0:
                    print(
                        "epoch {} iteration {} : batch.shape = {}, cost_time = {}us"
                        .format(e, i, batch.shape, cost_time0))
    world_rank = get_rank()
    tensor = torch.ones(2, 16) * world_rank
    print(
        f"[{os.getpid()}] : world_rank{world_rank} start all_reduce after dataloading"
    )
    dist.all_reduce(tensor, op=dist.ReduceOp.SUM, async_op=False)
    dist.barrier()
    print(
        f"[{os.getpid()}] : world_rank{world_rank} finished all_reduce after dataloading"
    )


class KSpeedDatasetWithChecking(KSpeedDataset):

    def __init__(self,
                 root,
                 object_name=None,
                 dl_workers=0,
                 device='CPU',
                 batchsize=1,
                 kspeed_iplist=None,
                 username='admin',
                 password='admin'):
        super(KSpeedDatasetWithChecking,
              self).__init__(root, object_name, dl_workers, device, batchsize,
                             kspeed_iplist, username, password)
        dataset_npy = np.load(self.root, mmap_mode='r')
        self.dataset_npy = dataset_npy.reshape(-1, dataset_npy.shape[-1])

    def __getitem__(self, idx):
        sample = None
        try:
            if self.dataset is None:
                self.load()

            if not isinstance(idx, int):
                idx_a = np.array(idx)
                sample = self.dataset.getbatch(idx_a)
                self.getbatch_check(self.device, sample, idx_a)
            else:
                sample = self.dataset.get(idx)
                self.get_check(self.device, sample, idx)
        except Exception as e:
            print("ERROR: kspeed read failed : {}".format(e))
            sys.exit()

        if sample is None:
            print("ERROR: kspeed read miss, idx : {}".format(idx))

        return sample

    def get_check(self, device, data, idx):
        if device == 'CPU':
            npt.assert_array_equal(data, self.dataset_npy[idx])
        else:
            import torch
            src_tensor = torch.from_numpy(self.dataset_npy[idx]).cuda()
            torch.cuda.synchronize()
            data_tensor = torch.as_tensor(data, device=torch.device('cuda'))
            torch.equal(data_tensor, src_tensor)

    def getbatch_check(self, device, data, idx_a):
        if device == 'CPU':
            npt.assert_array_equal(data, self.dataset_npy[idx_a])
        else:
            src_tensor = torch.from_numpy(self.dataset_npy[idx_a]).cuda()
            torch.cuda.synchronize()
            data_tensor = torch.as_tensor(data, device=torch.device('cuda'))
            torch.equal(data_tensor, src_tensor)


class KSpeedFolderWithChecking(KSpeedFolder):

    def __init__(self,
                 root,
                 object_name=None,
                 dl_workers=0,
                 kspeed_iplist=None,
                 username='admin',
                 password='admin'):
        super(KSpeedFolderWithChecking,
              self).__init__(root, object_name, dl_workers, kspeed_iplist,
                             username, password)

    def __getitem__(self, idx):
        if self.dataset is None:
            self.load()

        path, target = self.files[idx]  #获取图像（路径，标签）
        data_out = self.dataset.getBIN(path)  #数据读取
        if data_out is None:
            print("KSpeed Read dataout is None")
        else:
            with open(path, 'rb') as f:
                src_data = f.read()
                if data_out is not None:
                    if len(src_data) != len(data_out):
                        print(path + " check failed!!!!! length not same!")

                    if data_out != src_data:
                        print(path + " check failed!!!!!")

        return data_out, target


def collate_fake(batch):
    if (len(batch[0]) == 1):
        return batch
    imgs = [img[0] for img in batch]
    targets = torch.tensor([target[1] for target in batch], dtype=torch.int64)
    tensor = torch.zeros((len(imgs), 3, 244, 244), dtype=torch.uint8)
    return tensor, targets


def main_worker(gpuid, args):
    setup_training(gpuid, args.ngpus)
    #training dataset
    if os.path.isdir(args.root):
        if not args.check:
            print("In KSpeedFolder")
            train_ds = KSpeedFolder(args.root, args.objname, args.dlworker,
                                    args.iplist, args.username,
                                    args.password)
        else:
            print("In KSpeedFolderWithChecking")
            train_ds = KSpeedFolderWithChecking(args.root, args.objname,
                                                args.dlworker, args.iplist,
                                                args.username,
                                                args.password)
    else:
        if not args.check:
            print("In KSpeedDataset")
            train_ds = KSpeedDataset(args.root, args.objname,
                                        args.dlworker, args.device,
                                        args.batch_size, args.iplist,
                                        args.username, args.password)
        else:
            print("In KSpeedDatasetWithChecking")
            train_ds = KSpeedDatasetWithChecking(
                args.root, args.objname, args.dlworker, args.device,
                args.batch_size, args.iplist, args.username, args.password)

    collate_fn = collate_fake

    #training dataloader
    if args.batch_mode == 'None':
        if torch.distributed.is_initialized():
            train_sampler = torch.utils.data.distributed.DistributedSampler(
                train_ds)
        else:
            train_sampler = None

        dataloader = torch.utils.data.DataLoader(train_ds,
                                                 batch_size=args.batch_size,
                                                 sampler=train_sampler,
                                                 collate_fn=collate_fn,
                                                 num_workers=args.dlworker)
    else:
        if args.shuffle:
            batch_sampler_tr = BatchSampler(RandomSampler(train_ds),
                                            batch_size=args.batch_size,
                                            drop_last=True)
        else:
            batch_sampler_tr = BatchSampler(SequentialSampler(train_ds),
                                            batch_size=args.batch_size,
                                            drop_last=True)

        dataloader = torch.utils.data.DataLoader(train_ds,
                                                sampler=batch_sampler_tr,
                                                batch_size=None,
                                                num_workers=args.dlworker)

    #data loading test
    data_loading(dataloader, args)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i',
                        '--iplist',
                        dest="iplist",
                        default=None,
                        type=str,
                        help='iplist of kspeed server')
    parser.add_argument('-b',
                        '--batch_size',
                        dest="batch_size",
                        default=256,
                        type=int,
                        help='batch_size for distributed training')
    parser.add_argument('-n',
                        '--dlworker',
                        dest="dlworker",
                        default=4,
                        type=int,
                        help='dlworker for dataloader')
    parser.add_argument('-e',
                        '--epoch_num',
                        dest="epoch_num",
                        default=1,
                        type=int,
                        help='epoch_num for distributed training')
    parser.add_argument('-u',
                        "--username",
                        type=str,
                        dest="username",
                        default='admin',
                        help="specify the username to login")
    parser.add_argument('-w',
                        "--password",
                        type=str,
                        dest="password",
                        default='admin',
                        help="specify the password to login")
    parser.add_argument('-m',
                        '--batch_mode',
                        dest="batch_mode",
                        default='None',
                        type=str,
                        choices=["batch", "None"],
                        help='specify the mode using to load a batch of data ')
    parser.add_argument('-g',
                        '--ngpus',
                        dest="ngpus",
                        default=1,
                        type=int,
                        help='gpu num to use ')
    parser.add_argument('-o',
                        "--objname",
                        dest="objname",
                        type=str,
                        help="specify the kspeed object name to test")
    parser.add_argument('-r',
                        "--root",
                        dest="root",
                        type=str,
                        help="specify the root path of dataset")
    parser.add_argument("--shuffle",
                        dest="shuffle",
                        type=bool,
                        default=True,
                        help="specify True to have the data shuffled")
    parser.add_argument('-d',
                        '--device',
                        choices=['CPU', 'GPU'],
                        default='CPU',
                        help="Specify the device to get data to.")
    parser.add_argument('-c',
                        '--check',
                        type=bool,
                        default=False,
                        help="Whether to check the data get from kspeed.")

    args = parser.parse_args()
    if args.check is True and args.shuffle is True:
        #raise ValueError('check option is mutually exclusive with shuffle')
        print(
            'Check option is mutually exclusive with shuffle, forcibly set shuffle to False'
        )
        args.shuffle = False
    print("COMMAND: {}".format(args))
    return args


def main():
    args = parse_args()
    if os.getenv('gpus') is not None:
        args.ngpus = int(os.environ['gpus'])
    mp.spawn(main_worker, nprocs=args.ngpus, args=(args, ))


if __name__ == '__main__':
    main()
